from crosswalk_client.exceptions import MalformedThreshold


def validate_threshold_kwarg(function):
    def wrapper(*args, **kwargs):
        client = args[0]
        threshold = kwargs.get("threshold", client.threshold)
        if not isinstance(threshold, int):
            raise MalformedThreshold("Threshold should be an integer")
        if threshold < 0 or threshold > 100:
            raise MalformedThreshold("Threshold should be between 0 and 100")
        return function(*args, **kwargs)

    return wrapper



import uuid

import us

import pytest
from crosswalk_client import Client
from crosswalk_client.exceptions import (
    BadResponse,
    CreateEntityError,
    MalformedUpdateAttributes,
    UnspecificQueryError,
    EntityNotFoundError,
)


def test_setup(token, service):
    client = Client(token, service)
    client.create_domain("states")


def test_create(token, service):
    client = Client(token, service)
    domain = client.get_domain("states")
    client.set_domain(domain)
    entity = client.create({"name": "District of Columbia"})
    assert entity.name == "District of Columbia"
    entity.delete()  # Taxation w/out representation!


def test_bulk_create(token, service):
    client = Client(token, service)
    domain = client.get_domain("states")
    states = [
        {"name": s.name, "postal_code": s.abbr, "country": "USA"}
        for s in us.states.STATES
    ]
    entities = client.bulk_create(states, domain=domain)
    assert entities[0].name == states[0]["name"]


def test_bulk_create_nested_entity_error(token, service):
    client = Client(token, service, domain="states")
    entities = [{"a": "bad entity", "nested": {"too": "deeply"}}]
    with pytest.raises(CreateEntityError):
        client.bulk_create(entities)


def test_bulk_create_entity_with_reserved_attribute_error(token, service):
    client = Client(token, service, domain="states")
    entities = [
        {"a": "bad entity", "created": "with a reserved key, 'created'"}
    ]
    with pytest.raises(CreateEntityError):
        client.bulk_create(entities)


def test_bulk_create_entity_with_uuid(token, service):
    client = Client(token, service, domain="states")
    an_uuid = uuid.uuid4()
    entities = [{"uuid": an_uuid, "name": "an entity with a uuid"}]
    entity = client.bulk_create(entities)[0]
    assert entity.uuid == an_uuid
    entity.delete()


def test_entity_attributes_of_various_types(token, service):
    client = Client(token, service, domain="states")
    entities = [{"real": False, "number": 53, "listicle": [1, "one"]}]
    entity = client.bulk_create(entities)[0]
    assert entity.real is False
    assert entity.number == 53
    assert entity.listicle == [1, "one"]
    entity.delete()


def test_best_match(token, service):
    client = Client(token, service, domain="states")
    entity = client.best_match({"name": "Misisipi"})
    assert entity.name == "Mississippi"


def test_best_match_with_block_attrs(token, service):
    client = Client(token, service, domain="states")

    entity = client.best_match(
        {"name": "Arkansas"}, block_attrs={"postal_code": "KS"}
    )
    assert entity.name == "Kansas"

    entity = client.best_match({"name": "Arkansas"})
    assert entity.name == "Arkansas"


def test_match(token, service):
    client = Client(token, service, domain="states")
    entity = client.match(
        {"name": "Kansas"}, block_attrs={"postal_code": "KS"}
    )
    assert entity.name == "Kansas"


def test_match_exceptions(token, service):
    client = Client(token, service, domain="states")
    with pytest.raises(UnspecificQueryError):
        client.match({"country": "USA"})
    with pytest.raises(EntityNotFoundError):
        client.match({"country": "Deutschland"})


def test_match_or_create(token, service):
    client = Client(token, service, domain="states")
    entity = client.match_or_create(
        {"name": "Thunderdome"}, create_attrs={"country": "Australia"}
    )
    assert entity.created is True and entity.name == "Thunderdome"

    entity2 = client.match_or_create(
        {"name": "Gas City"}, create_attrs={"country": "Australia"}
    )
    assert entity2.created is True and entity2.name == "Gas City"

    with pytest.raises(UnspecificQueryError):
        client.match_or_create({"country": "Australia"})

    entity.delete()
    entity2.delete()


def test_best_match_or_create(token, service):
    client = Client(token, service, domain="states")
    entity = client.best_match_or_create({"name": "Narnia"}, threshold=75)
    assert entity.created is True and entity.name == "Narnia"


def test_best_match_or_create_with_uuid(token, service):
    client = Client(token, service, domain="states")
    an_uuid = uuid.uuid4()
    entity = client.best_match_or_create(
        {"name": "Xanadu"}, create_attrs={"uuid": an_uuid}, threshold=75
    )
    assert entity.uuid == an_uuid


def test_get_entities(token, service):
    client = Client(token, service)
    states = client.get_domain("states")
    entities = client.get_entities(domain=states)
    assert len(entities) > 50
    assert entities[0].name is not None


def test_get_entity(token, service):
    client = Client(token, service, domain="states")
    domain = client.get_domain("states")
    entity = client.get_entities(domain=domain)[0]
    returned_entity = client.get_entity(entity.uuid)
    assert returned_entity.uuid == entity.uuid


def test_get_entities_with_block_attrs(token, service):
    client = Client(token, service, domain="states")
    entities = client.get_entities({"country": "USA"})
    assert len(entities) == 51

    entities = client.get_entities({"postal_code": "KS"})
    assert entities[0].name == "Kansas"


def test_update_entity_by_id(token, service):
    client = Client(token, service, domain="states")
    entity = client.best_match({"name": "Xanadu"})
    entity = client.update_by_id(
        entity.uuid, {"sacred river": "Alph", "name": "Zanadu"}
    )
    assert entity.name == "Zanadu"
    entity = client.update_by_id(entity.uuid, {"name": "Xanadu"})
    assert entity.sacred_river == "Alph" and entity.name == "Xanadu"


def test_update_entity_by_match(token, service):
    client = Client(token, service, domain="states")
    entity = client.update_match(
        {"name": "Xanadu"}, {"sacred river": "Mississippi"}
    )
    assert entity.sacred_river == "Mississippi"


def test_update_entity_by_match_error(token, service):
    client = Client(token, service, domain="states")
    with pytest.raises(UnspecificQueryError):
        client.update_match(
            {"country": "USA"}, {"sacred river": "Mississippi"}
        )


def test_update_entity_by_match_invalid_data_error(token, service):
    client = Client(token, service, domain="states")
    with pytest.raises(MalformedUpdateAttributes):
        client.update_match(
            {"name": "Xanadu"}, {"stuff": {"nested": "too deeply"}}
        )


def test_delete_entity_by_match(token, service):
    client = Client(token, service, domain="states")
    deleted = client.delete_match({"name": "Narnia"})
    assert deleted is True


def test_delete_entity_by_id(token, service):
    client = Client(token, service, domain="states")
    entity = client.best_match({"name": "Xanadu"})
    response = client.delete_by_id(entity.uuid)
    assert response is True


def test_unspecific_delete_error(token, service):
    client = Client(token, service, domain="states")
    with pytest.raises(UnspecificQueryError):
        client.delete_match({"country": "USA"})


def test_bad_domain_delete_error(token, service):
    client = Client(token, service)
    client.set_domain("countries")  # does not exist
    with pytest.raises(BadResponse):
        client.delete_match({"country": "USA"})


def test_create_or_alias_creates_alias(token, service):
    client = Client(token, service, domain="states")
    entity = client.alias_or_create({"name": "Kalifornia"})
    assert entity.name == "California" and entity.aliased is True


def test_create_or_alias_creates_new(token, service):
    client = Client(token, service, domain="states")
    entity = client.alias_or_create(
        {"name": "Alderaan"},
        create_attrs={"galaxy": "Far, far away"},
        threshold=90,
    )
    assert entity.name == "Alderaan" and entity.aliased is False


def test_create_or_alias_fails_on_existing_entity(token, service):
    client = Client(token, service, domain="states")
    with pytest.raises(BadResponse):
        client.alias_or_create(
            {"name": "Alderaan"}, create_attrs={"galaxy": "Far, far away"}
        )


def test_best_match_an_alias(token, service):
    client = Client(token, service, domain="states")
    entity = client.best_match({"name": "Kalifornia"})
    assert entity.name == "California"


def test_entity_cannot_alias_itself(token, service):
    client = Client(token, service)
    domain = client.create_domain("temp domain")
    entity1, entity2 = client.bulk_create(
        [{"name": "A fake Entity"}, {"name": "A second faker"}], domain=domain
    )
    with pytest.raises(BadResponse):
        entity1.set_alias_for(entity1)
    assert entity1.alias_for is None
    entity1.set_alias_for(entity2)
    assert entity1.alias_for == entity2.uuid
    entity1.delete()
    entity2.delete()
    domain.delete()


def test_empty_domain(token, service):
    client = Client(token, service)
    domain = client.create_domain("empty domain")
    entity = client.best_match({"name": "does not exist"}, domain=domain)
    assert entity is None

    entity = client.best_match_or_create(
        {"name": "should create"}, domain=domain
    )
    assert entity.created is True
    assert entity.name == "should create"

    entity.delete()
    domain.delete()


def test_cleanup(token, service):
    client = Client(token, service, domain="states")
    for state in us.states.STATES:
        client.delete_match({"name": state.name})
    client.delete_match({"name": "Alderaan"})
    client.delete_domain("states")



from urllib.parse import urljoin

import requests

from crosswalk_client.exceptions import BadResponse
from crosswalk_client.objects.domain import DomainObject
from crosswalk_client.validators.domain import (
    validate_required_domain_string_arg
)


class GetDomain(object):
    @validate_required_domain_string_arg
    def get_domain(self, domain):
        response = requests.get(
            urljoin(self.service_address, f"domains/{domain}/"),
            headers=self.headers,
        )
        if response.status_code != requests.codes.ok:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        return DomainObject(response.json(), client=self)



from crosswalk_client.validators.domain import validate_required_domain_arg


class SetDomain(object):
    @validate_required_domain_arg
    def set_domain(self, domain):
        self.domain = domain



from crosswalk_client.exceptions import MalformedDomain, MissingDomain


def validate_required_domain_string_arg(function):
    """
    Validates a domain string name is passed.
    """

    def wrapper(*args, **kwargs):
        list_args = list(args)
        try:
            domain = list_args[1]
        except IndexError:
            raise MissingDomain("Must pass a domain")
        if not isinstance(domain, str):
            raise MalformedDomain("Domain must be a string or domain instance")
        return function(*args, **kwargs)

    return wrapper



from crosswalk_client.exceptions import MalformedUpdateAttributes


def validate_update_attrs_kwarg(function):
    def wrapper(*args, **kwargs):
        update_attrs = kwargs.get("update_attrs", {}).copy()
        if not isinstance(update_attrs, dict):
            raise MalformedUpdateAttributes(
                "Update attributes must be a dictionary"
            )
        for key in update_attrs:
            if isinstance(update_attrs[key], dict):
                raise MalformedUpdateAttributes(
                    "Nested attributes are not allowed in update attributes"
                )
        return function(*args, **kwargs)

    return wrapper



from urllib.parse import urljoin

import requests

from crosswalk_client.encoder import encode
from crosswalk_client.exceptions import (
    BadResponse,
    UnspecificQueryError,
    UpdateEntityError,
)
from crosswalk_client.objects.entity import EntityObject
from crosswalk_client.validators.entity import (
    validate_domain_kwarg,
    validate_required_block_attrs_arg,
    validate_required_update_attrs_arg,
)


class UpdateMatch(object):
    """
    Update a match.

    Entites should be an array of attributes dicts.
    """

    @validate_required_block_attrs_arg
    @validate_required_update_attrs_arg
    @validate_domain_kwarg
    def update_match(self, block_attrs, update_attrs, domain=None):
        if domain is None:
            domain = self.domain
        data = {"block_attrs": block_attrs, "update_attrs": update_attrs}
        response = requests.post(
            urljoin(
                self.service_address,
                "domains/{}/entities/update-match/".format(domain),
            ),
            headers=self.headers,
            data=encode(data),
        )
        if response.status_code == 403:
            raise UnspecificQueryError(
                "Query in request matched more than one entity."
            )
        if response.status_code == 400:
            raise UpdateEntityError(response.content)
        if response.status_code != requests.codes.ok:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        return EntityObject(response.json(), client=self)



from urllib.parse import urljoin

import requests

from crosswalk_client.exceptions import BadResponse
from crosswalk_client.objects.domain import DomainObject
from crosswalk_client.validators.domain import (
    validate_parent_domain_kwarg,
    validate_required_domain_string_arg,
)


class CreateDomain(object):
    """ Create a domain. """

    @validate_parent_domain_kwarg
    @validate_required_domain_string_arg
    def create_domain(self, domain, parent=None):
        if parent:
            data = {"name": domain, "parent": parent}
        else:
            data = {"name": domain}
        response = requests.post(
            urljoin(self.service_address, "domains/"),
            headers=self.headers,
            json=data,
        )
        if response.status_code != 201:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        return DomainObject(response.json(), client=self)



from slugify import slugify

from crosswalk_client.exceptions import MalformedDomain, MissingDomain
from crosswalk_client.objects.domain import DomainObject


def validate_domain_kwarg(function):
    def wrapper(*args, **kwargs):
        client = args[0]
        domain = kwargs.get("domain", client.domain)
        if domain is None:
            raise MissingDomain("Must set a domain")
        if isinstance(domain, DomainObject):
            if "domain" in kwargs:
                kwargs["domain"] = domain.slug
            args[0].domain = domain.slug
            return function(*args, **kwargs)
        if not isinstance(domain, str):
            raise MalformedDomain("Domain must be a string or domain instance")
        if slugify(domain) != domain:
            raise MalformedDomain("Domain string should be a slug")
        return function(*args, **kwargs)

    return wrapper



from urllib.parse import urljoin

import requests

from crosswalk_client.encoder import encode
from crosswalk_client.exceptions import BadResponse, CreateEntityError
from crosswalk_client.objects.entity import EntityObject
from crosswalk_client.validators.entity import validate_domain_kwarg


class Create(object):
    """
    Create a single entity. (Just a shortcut to bulk create...)

    Entity should be a dict of attributes.
    """

    @validate_domain_kwarg
    def create(self, entity, domain=None):
        if domain is None:
            domain = self.domain
        response = requests.post(
            urljoin(
                self.service_address,
                "domains/{}/entities/bulk-create/".format(domain),
            ),
            headers=self.headers,
            data=encode([entity]),
        )
        if response.status_code == 400:
            raise CreateEntityError(
                "Error creating entity: {}".format(response.content)
            )
        if response.status_code != requests.codes.ok:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        return [
            EntityObject(entity, client=self)
            for entity in response.json()["entities"]
        ][0]



class DomainObject(object):
    """
    Converts json returned form server into a proper domain object.
    """

    def update(self, update_attrs):
        domain = self.__client.update_domain(self.slug, update_attrs)
        self.__dict__ = domain.__dict__

    def set_parent(self, parent):
        domain = self.__client.update_domain(
            self.slug, {"parent": parent.slug}
        )
        self.__dict__ = domain.__dict__

    def remove_parent(self):
        domain = self.__client.update_domain(self.slug, {"parent": None})
        self.__dict__ = domain.__dict__

    def delete(self):
        deleted = self.__client.delete_domain(self.slug)
        if deleted:
            self.__dict__ = {"slug": self.slug, "deleted": deleted}

    def __init__(self, response, client=None):
        self.__dict__ = response
        self.__client = client

    def __repr__(self):
        return f"<Crosswalk Domain: {self.slug}>"
